import { getParsedType, hasKeys, INVALID, isOneOf, OK, ParsedType } from "@postgresql-typed/util";
import { getBigNumber } from "../../util/getBigNumber.js";
import { PGTPBase } from "../../util/PGTPBase.js";
import { PGTPConstructorBase } from "../../util/PGTPConstructorBase.js";
const bigNumber = getBigNumber("-92233720368547758.08", "92233720368547758.07");
class MoneyConstructorClass extends PGTPConstructorBase {
    constructor() {
        super();
    }
    _parse(context) {
        if (context.data.length !== 1) {
            this.setIssueForContext(context, context.data.length > 1
                ? {
                    code: "too_big",
                    type: "arguments",
                    maximum: 1,
                    exact: true,
                }
                : {
                    code: "too_small",
                    type: "arguments",
                    minimum: 1,
                    exact: true,
                });
            return INVALID;
        }
        const [argument] = context.data, allowedTypes = [ParsedType.number, ParsedType.string, ParsedType.object, ParsedType.bigNumber, ParsedType.bigint], parsedType = getParsedType(argument);
        if (!isOneOf(allowedTypes, parsedType)) {
            this.setIssueForContext(context, {
                code: "invalid_type",
                expected: allowedTypes,
                received: parsedType,
            });
            return INVALID;
        }
        switch (parsedType) {
            case ParsedType.object:
                return this._parseObject(context, argument);
            default:
                return this._parseString(context, argument);
        }
    }
    _parseString(context, argument) {
        const parsedNumber = bigNumber(argument);
        if (parsedNumber.success)
            return OK(new MoneyClass(parsedNumber.data));
        this.setIssueForContext(context, parsedNumber.error);
        return INVALID;
    }
    _parseObject(context, argument) {
        if (this.isMoney(argument))
            return OK(new MoneyClass(argument.money));
        const parsedObject = hasKeys(argument, [["value", "string"]]);
        if (parsedObject.success)
            return this._parseString(context, parsedObject.obj.value);
        switch (true) {
            case parsedObject.otherKeys.length > 0:
                this.setIssueForContext(context, {
                    code: "unrecognized_keys",
                    keys: parsedObject.otherKeys,
                });
                break;
            case parsedObject.missingKeys.length > 0:
                this.setIssueForContext(context, {
                    code: "missing_keys",
                    keys: parsedObject.missingKeys,
                });
                break;
            case parsedObject.invalidKeys.length > 0:
                this.setIssueForContext(context, {
                    code: "invalid_key_type",
                    ...parsedObject.invalidKeys[0],
                });
                break;
        }
        return INVALID;
    }
    isMoney(object) {
        return object instanceof MoneyClass;
    }
}
const Money = new MoneyConstructorClass();
class MoneyClass extends PGTPBase {
    constructor(_money) {
        super();
        this._money = _money;
    }
    _equals(input) {
        //@ts-expect-error - _equals receives the same context as _parse
        const parsed = Money.safeFrom(...input.data);
        if (parsed.success) {
            return OK({
                equals: parsed.data.toString() === this.toString(),
                data: parsed.data,
            });
        }
        this.setIssueForContext(input, parsed.error.issue);
        return INVALID;
    }
    toString() {
        const money = this._money.toString();
        //* Should always have a decimal point, with 2 decimal places
        if (money.includes(".")) {
            const [int, decimal] = money.split(".");
            //* If it has more than 2 decimal places, round it
            if (decimal.length > 2)
                return `${int}.${Math.round(Number(`0.${decimal}`) * 100)}`;
            return `${int}.${decimal.padEnd(2, "0")}`;
        }
        else
            return `${money}.00`;
    }
    toBigNumber() {
        return this._money;
    }
    toJSON() {
        return {
            value: this.toString(),
        };
    }
    get money() {
        return this._money;
    }
    set money(money) {
        const parsed = Money.safeFrom(money);
        if (parsed.success)
            this._money = parsed.data.toBigNumber();
        else
            throw parsed.error;
    }
    get value() {
        return this.toString();
    }
    set value(money) {
        const parsed = Money.safeFrom(money);
        if (parsed.success)
            this._money = parsed.data.money;
        else
            throw parsed.error;
    }
    get postgres() {
        return this.toString();
    }
    set postgres(money) {
        const parsed = Money.safeFrom(money);
        if (parsed.success)
            this._money = parsed.data.money;
        else
            throw parsed.error;
    }
}
export { Money };
//# sourceMappingURL=Money.js.map