import { getParsedType, hasKeys, INVALID, isOneOf, OK, ParsedType } from "@postgresql-typed/util";
import { getBigNumber } from "../../util/getBigNumber.js";
import { PGTPBase } from "../../util/PGTPBase.js";
import { PGTPConstructorBase } from "../../util/PGTPConstructorBase.js";
const bigNumber = getBigNumber("-99999999999999999999999999999999999999", "99999999999999999999999999999999999999", {
    allowInfinity: true,
    allowNaN: true,
});
class Float4ConstructorClass extends PGTPConstructorBase {
    constructor() {
        super();
    }
    _parse(context) {
        if (context.data.length !== 1) {
            this.setIssueForContext(context, context.data.length > 1
                ? {
                    code: "too_big",
                    type: "arguments",
                    maximum: 1,
                    exact: true,
                }
                : {
                    code: "too_small",
                    type: "arguments",
                    minimum: 1,
                    exact: true,
                });
            return INVALID;
        }
        const [argument] = context.data, allowedTypes = [ParsedType.number, ParsedType.string, ParsedType.object, ParsedType.nan, ParsedType.infinity, ParsedType.bigNumber, ParsedType.bigint], parsedType = getParsedType(argument);
        if (!isOneOf(allowedTypes, parsedType)) {
            this.setIssueForContext(context, {
                code: "invalid_type",
                expected: allowedTypes,
                received: parsedType,
            });
            return INVALID;
        }
        switch (parsedType) {
            case ParsedType.object:
                return this._parseObject(context, argument);
            default:
                return this._parseString(context, argument);
        }
    }
    _parseString(context, argument) {
        const parsedNumber = bigNumber(argument);
        if (parsedNumber.success)
            return OK(new Float4Class(parsedNumber.data));
        this.setIssueForContext(context, parsedNumber.error);
        return INVALID;
    }
    _parseObject(context, argument) {
        if (this.isFloat4(argument))
            return OK(new Float4Class(argument.float4));
        const parsedObject = hasKeys(argument, [["value", "string"]]);
        if (parsedObject.success)
            return this._parseString(context, parsedObject.obj.value);
        switch (true) {
            case parsedObject.otherKeys.length > 0:
                this.setIssueForContext(context, {
                    code: "unrecognized_keys",
                    keys: parsedObject.otherKeys,
                });
                break;
            case parsedObject.missingKeys.length > 0:
                this.setIssueForContext(context, {
                    code: "missing_keys",
                    keys: parsedObject.missingKeys,
                });
                break;
            case parsedObject.invalidKeys.length > 0:
                this.setIssueForContext(context, {
                    code: "invalid_key_type",
                    ...parsedObject.invalidKeys[0],
                });
                break;
        }
        return INVALID;
    }
    isFloat4(object) {
        return object instanceof Float4Class;
    }
}
const Float4 = new Float4ConstructorClass();
class Float4Class extends PGTPBase {
    constructor(_float4) {
        super();
        this._float4 = _float4;
    }
    _equals(input) {
        //@ts-expect-error - _equals receives the same context as _parse
        const parsed = Float4.safeFrom(...input.data);
        if (parsed.success) {
            return OK({
                equals: parsed.data.toString() === this.toString(),
                data: parsed.data,
            });
        }
        this.setIssueForContext(input, parsed.error.issue);
        return INVALID;
    }
    toString() {
        return this._float4.toString();
    }
    toBigNumber() {
        return this._float4;
    }
    toJSON() {
        return {
            value: this._float4.toString(),
        };
    }
    get float4() {
        return this._float4;
    }
    set float4(float4) {
        const parsed = Float4.safeFrom(float4);
        if (parsed.success)
            this._float4 = parsed.data.toBigNumber();
        else
            throw parsed.error;
    }
    get value() {
        return this.toString();
    }
    set value(float4) {
        const parsed = Float4.safeFrom(float4);
        if (parsed.success)
            this._float4 = parsed.data.float4;
        else
            throw parsed.error;
    }
    get postgres() {
        return this.toString();
    }
    set postgres(float4) {
        const parsed = Float4.safeFrom(float4);
        if (parsed.success)
            this._float4 = parsed.data.float4;
        else
            throw parsed.error;
    }
}
export { Float4 };
//# sourceMappingURL=Float4.js.map