import { getParsedType, hasKeys, INVALID, isOneOf, OK, ParsedType } from "@postgresql-typed/util";
import { getBigNumber } from "../../util/getBigNumber.js";
import { PGTPBase } from "../../util/PGTPBase.js";
import { PGTPConstructorBase } from "../../util/PGTPConstructorBase.js";
const bigNumber = getBigNumber("-10e307", "10e307", {
    allowInfinity: true,
    allowNaN: true,
});
class Float8ConstructorClass extends PGTPConstructorBase {
    constructor() {
        super();
    }
    _parse(context) {
        if (context.data.length !== 1) {
            this.setIssueForContext(context, context.data.length > 1
                ? {
                    code: "too_big",
                    type: "arguments",
                    maximum: 1,
                    exact: true,
                }
                : {
                    code: "too_small",
                    type: "arguments",
                    minimum: 1,
                    exact: true,
                });
            return INVALID;
        }
        const [argument] = context.data, allowedTypes = [ParsedType.number, ParsedType.string, ParsedType.object, ParsedType.nan, ParsedType.infinity, ParsedType.bigNumber, ParsedType.bigint], parsedType = getParsedType(argument);
        if (!isOneOf(allowedTypes, parsedType)) {
            this.setIssueForContext(context, {
                code: "invalid_type",
                expected: allowedTypes,
                received: parsedType,
            });
            return INVALID;
        }
        switch (parsedType) {
            case ParsedType.object:
                return this._parseObject(context, argument);
            default:
                return this._parseString(context, argument);
        }
    }
    _parseString(context, argument) {
        const parsedNumber = bigNumber(argument);
        if (parsedNumber.success)
            return OK(new Float8Class(parsedNumber.data));
        this.setIssueForContext(context, parsedNumber.error);
        return INVALID;
    }
    _parseObject(context, argument) {
        if (this.isFloat8(argument))
            return OK(new Float8Class(argument.float8));
        const parsedObject = hasKeys(argument, [["value", "string"]]);
        if (parsedObject.success)
            return this._parseString(context, parsedObject.obj.value);
        switch (true) {
            case parsedObject.otherKeys.length > 0:
                this.setIssueForContext(context, {
                    code: "unrecognized_keys",
                    keys: parsedObject.otherKeys,
                });
                break;
            case parsedObject.missingKeys.length > 0:
                this.setIssueForContext(context, {
                    code: "missing_keys",
                    keys: parsedObject.missingKeys,
                });
                break;
            case parsedObject.invalidKeys.length > 0:
                this.setIssueForContext(context, {
                    code: "invalid_key_type",
                    ...parsedObject.invalidKeys[0],
                });
                break;
        }
        return INVALID;
    }
    isFloat8(object) {
        return object instanceof Float8Class;
    }
}
const Float8 = new Float8ConstructorClass();
class Float8Class extends PGTPBase {
    constructor(_float8) {
        super();
        this._float8 = _float8;
    }
    _equals(input) {
        //@ts-expect-error - _equals receives the same context as _parse
        const parsed = Float8.safeFrom(...input.data);
        if (parsed.success) {
            return OK({
                equals: parsed.data.toString() === this.toString(),
                data: parsed.data,
            });
        }
        this.setIssueForContext(input, parsed.error.issue);
        return INVALID;
    }
    toString() {
        return this._float8.toString();
    }
    toBigNumber() {
        return this._float8;
    }
    toJSON() {
        return {
            value: this._float8.toString(),
        };
    }
    get float8() {
        return this._float8;
    }
    set float8(float8) {
        const parsed = Float8.safeFrom(float8);
        if (parsed.success)
            this._float8 = parsed.data.toBigNumber();
        else
            throw parsed.error;
    }
    get value() {
        return this.toString();
    }
    set value(float8) {
        const parsed = Float8.safeFrom(float8);
        if (parsed.success)
            this._float8 = parsed.data.float8;
        else
            throw parsed.error;
    }
    get postgres() {
        return this.toString();
    }
    set postgres(float8) {
        const parsed = Float8.safeFrom(float8);
        if (parsed.success)
            this._float8 = parsed.data.float8;
        else
            throw parsed.error;
    }
}
export { Float8 };
//# sourceMappingURL=Float8.js.map