import { getParsedType, hasKeys, INVALID, isOneOf, OK, ParsedType } from "@postgresql-typed/util";
import { PGTPBase } from "../../util/PGTPBase.js";
import { PGTPConstructorBase } from "../../util/PGTPConstructorBase.js";
class Int2ConstructorClass extends PGTPConstructorBase {
    constructor() {
        super();
    }
    _parse(context) {
        if (context.data.length !== 1) {
            this.setIssueForContext(context, context.data.length > 1
                ? {
                    code: "too_big",
                    type: "arguments",
                    maximum: 1,
                    exact: true,
                }
                : {
                    code: "too_small",
                    type: "arguments",
                    minimum: 1,
                    exact: true,
                });
            return INVALID;
        }
        const [argument] = context.data, allowedTypes = [ParsedType.number, ParsedType.string, ParsedType.object], parsedType = getParsedType(argument);
        if (!isOneOf(allowedTypes, parsedType)) {
            this.setIssueForContext(context, {
                code: "invalid_type",
                expected: allowedTypes,
                received: parsedType,
            });
            return INVALID;
        }
        switch (parsedType) {
            case ParsedType.number:
                return this._parseNumber(context, argument);
            case ParsedType.string:
                return this._parseString(context, argument);
            default:
                return this._parseObject(context, argument);
        }
    }
    _parseNumber(context, argument) {
        if (Number.isNaN(argument)) {
            this.setIssueForContext(context, {
                code: "invalid_type",
                expected: "number",
                received: "nan",
            });
            return INVALID;
        }
        if (!Number.isFinite(argument)) {
            this.setIssueForContext(context, {
                code: "not_finite",
            });
            return INVALID;
        }
        if (argument % 1 !== 0) {
            this.setIssueForContext(context, {
                code: "not_whole",
            });
            return INVALID;
        }
        if (argument < -32768) {
            this.setIssueForContext(context, {
                code: "too_small",
                type: "number",
                minimum: -32768,
                inclusive: true,
            });
            return INVALID;
        }
        if (argument > 32767) {
            this.setIssueForContext(context, {
                code: "too_big",
                type: "number",
                maximum: 32767,
                inclusive: true,
            });
            return INVALID;
        }
        return OK(new Int2Class(argument));
    }
    _parseString(context, argument) {
        const parsed = Number.parseFloat(argument);
        return this._parseNumber(context, parsed);
    }
    _parseObject(context, argument) {
        if (this.isInt2(argument))
            return OK(new Int2Class(argument.value));
        const parsedObject = hasKeys(argument, [["value", "number"]]);
        if (parsedObject.success)
            return this._parseNumber(context, parsedObject.obj.value);
        switch (true) {
            case parsedObject.otherKeys.length > 0:
                this.setIssueForContext(context, {
                    code: "unrecognized_keys",
                    keys: parsedObject.otherKeys,
                });
                break;
            case parsedObject.missingKeys.length > 0:
                this.setIssueForContext(context, {
                    code: "missing_keys",
                    keys: parsedObject.missingKeys,
                });
                break;
            case parsedObject.invalidKeys.length > 0:
                this.setIssueForContext(context, {
                    code: "invalid_key_type",
                    ...parsedObject.invalidKeys[0],
                });
                break;
        }
        return INVALID;
    }
    isInt2(object) {
        return object instanceof Int2Class;
    }
}
const Int2 = new Int2ConstructorClass();
class Int2Class extends PGTPBase {
    constructor(_int2) {
        super();
        this._int2 = _int2;
    }
    _equals(input) {
        //@ts-expect-error - _equals receives the same context as _parse
        const parsed = Int2.safeFrom(...input.data);
        if (parsed.success) {
            return OK({
                equals: parsed.data.toNumber() === this.toNumber(),
                data: parsed.data,
            });
        }
        this.setIssueForContext(input, parsed.error.issue);
        return INVALID;
    }
    toString() {
        return this._int2.toString();
    }
    toNumber() {
        return this._int2;
    }
    toJSON() {
        return {
            value: this._int2,
        };
    }
    get int2() {
        return this._int2;
    }
    set int2(int2) {
        const parsed = Int2.safeFrom(int2);
        if (parsed.success)
            this._int2 = parsed.data.toNumber();
        else
            throw parsed.error;
    }
    get value() {
        return this._int2;
    }
    set value(int2) {
        const parsed = Int2.safeFrom(int2);
        if (parsed.success)
            this._int2 = parsed.data.toNumber();
        else
            throw parsed.error;
    }
    get postgres() {
        return this.toString();
    }
    set postgres(int2) {
        const parsed = Int2.safeFrom(int2);
        if (parsed.success)
            this._int2 = parsed.data.toNumber();
        else
            throw parsed.error;
    }
}
export { Int2 };
//# sourceMappingURL=Int2.js.map